"""
Python module for the computation of maneuver sizing load cases.
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2023 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import fastoad.api as oad
import numpy as np
import openmdao.api as om
from openmdao.api import convert_units

from ..constants import SERVICE_MANEUVER_LOADS


@oad.RegisterSubmodel(SERVICE_MANEUVER_LOADS, "fastoad.submodel.loads.maneuver.legacy")
class ManeuverLoads(om.ExplicitComponent):
    """
    Computes CS25 pull-up maneuver load factors evaluated at two different load cases:

    Load case 1: with wings with almost no fuel
    Load case 2: at maximum take-off weight

    Based on formulas in :cite:`supaero:2014`, §6.3

    """

    def setup(self):
        self.add_input("data:load_case:maneuver_load_factor", val=0.0, units="unitless")
        self.add_input("data:mission:sizing:cs25:safety_factor", val=1.5, units="unitless")
        self.add_input("data:weight:aircraft:MTOW", val=np.nan, units="kg")

        self.add_output("data:mission:sizing:cs25:maneuver:load_factor_1", units="unitless")
        self.add_output("data:mission:sizing:cs25:maneuver:load_factor_2", units="unitless")

    def setup_partials(self):
        self.declare_partials(
            "data:mission:sizing:cs25:maneuver:load_factor_1",
            "*",
            method="fd",
        )
        self.declare_partials(
            "data:mission:sizing:cs25:maneuver:load_factor_2",
            "*",
            method="fd",
        )

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        n_maneuver = inputs["data:load_case:maneuver_load_factor"]
        safety_factor = inputs["data:mission:sizing:cs25:safety_factor"]
        mtow = inputs["data:weight:aircraft:MTOW"]

        if (
            n_maneuver == 0.0
        ):  # n_maneuver not provided by the user, so we evaluate it using CS 25.337
            n_maneuver = self.__n_maneuver(mtow)

        # load case #1
        n1 = safety_factor * n_maneuver

        # load case #2
        n2 = safety_factor * n_maneuver

        outputs["data:mission:sizing:cs25:maneuver:load_factor_1"] = n1
        outputs["data:mission:sizing:cs25:maneuver:load_factor_2"] = n2

    @staticmethod
    def __n_maneuver(w_kg):
        """
        See CS 25.337 Limit manoeuvring load factors.
        Computes the maneuver load factor based on gross weight (in kg).
        - Converts weight to pounds.
        - Uses the formula: 2 * (1 + 24000 / (w + 10000))
        - Clip the result between 2.5 and 3.8.
        """
        w = convert_units(w_kg, "kg", "lb")
        raw_n = 2 * (1 + 24000 / (w + 10000))
        return np.clip(raw_n, 2.5, 3.8)



"""Constants for CG submodels."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

SERVICE_HORIZONTAL_TAIL_CG = "service.cg.horizontal_tail"
SERVICE_VERTICAL_TAIL_CG = "service.cg.vertical_tail"
SERVICE_WING_CG = "service.cg.wing"
SERVICE_FLIGHT_CONTROLS_CG = "service.cg.wing.control_surfaces"
SERVICE_TANKS_CG = "service.cg.tanks"
SERVICE_OTHERS_CG = "service.cg.others"
SERVICE_GLOBAL_CG = "service.cg.global"
SERVICE_MLG_CG = "service.cg.mlg"
SERVICE_AIRCRAFT_CG = "service.cg.aircraft"
SERVICE_EMPTY_AIRCRAFT_CG = "service.cg.empty_aircraft"
SERVICE_LOAD_CASES_CG = "service.cg.load_cases"



"""
Estimation of horizontal tail lift coefficient
"""

#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2023 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om


# TODO: This belongs more to aerodynamics than geometry
class ComputeHTClalpha(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """Horizontal tail lift coefficient estimation"""

    def setup(self):
        self.add_input("data:geometry:horizontal_tail:aspect_ratio", val=np.nan)
        self.add_input("data:TLAR:cruise_mach", val=np.nan)
        self.add_input("data:geometry:horizontal_tail:sweep_25", val=np.nan, units="deg")

        self.add_output("data:aerodynamics:horizontal_tail:cruise:CL_alpha", units="1/rad")

    def setup_partials(self):
        self.declare_partials("data:aerodynamics:horizontal_tail:cruise:CL_alpha", "*", method="fd")

    def compute(self, inputs, outputs):
        cruise_mach = inputs["data:TLAR:cruise_mach"]
        lambda_ht = inputs["data:geometry:horizontal_tail:aspect_ratio"]
        sweep_25_ht = inputs["data:geometry:horizontal_tail:sweep_25"]

        beta = np.sqrt(1 - cruise_mach**2)
        cl_alpha_ht = (
            0.8
            * 2
            * np.pi
            * lambda_ht
            / (
                2
                + np.sqrt(
                    4
                    + lambda_ht**2
                    * beta**2
                    / 0.95**2
                    * (1 + (np.tan(sweep_25_ht / 180.0 * np.pi)) ** 2 / beta**2)
                )
            )
        )

        outputs["data:aerodynamics:horizontal_tail:cruise:CL_alpha"] = cl_alpha_ht



"""
Estimation of center wing chord (l1')
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2025 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from scipy.interpolate import make_interp_spline


class ComputeCenterChord(om.ExplicitComponent):
    """
    Estimation of center wing chord (l1') by extrapolation of the virtual edges of the wing.
    See Figure 3-3bis of :cite:`supaero:2014`.
    """

    def setup(self):
        self.add_input("data:geometry:wing:root:chord", val=np.nan, units="m")
        self.add_input("data:geometry:wing:root:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:kink:chord", val=np.nan, units="m")
        self.add_input("data:geometry:wing:kink:leading_edge:x:local", val=np.nan, units="m")
        self.add_input("data:geometry:wing:kink:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:tip:chord", val=np.nan, units="m")
        self.add_input("data:geometry:wing:tip:leading_edge:x:local", val=np.nan, units="m")
        self.add_input("data:geometry:wing:tip:y", val=np.nan, units="m")

        self.add_output("data:geometry:wing:center:chord", units="m")
        self.add_output("data:geometry:wing:center:leading_edge:x:local", units="m")

    def setup_partials(self):
        self.declare_partials(
            "data:geometry:wing:center:chord",
            [
                "data:geometry:wing:root:chord",
                "data:geometry:wing:root:y",
                "data:geometry:wing:kink:chord",
                "data:geometry:wing:kink:y",
                "data:geometry:wing:tip:chord",
                "data:geometry:wing:tip:leading_edge:x:local",
                "data:geometry:wing:tip:y",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:geometry:wing:center:leading_edge:x:local",
            [
                "data:geometry:wing:root:y",
                "data:geometry:wing:kink:leading_edge:x:local",
                "data:geometry:wing:kink:y",
                "data:geometry:wing:tip:leading_edge:x:local",
                "data:geometry:wing:tip:y",
            ],
            method="fd",
        )

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        l2_wing = inputs["data:geometry:wing:root:chord"]
        y2_wing = inputs["data:geometry:wing:root:y"]
        l3_wing = inputs["data:geometry:wing:kink:chord"]
        x3_wing = inputs["data:geometry:wing:kink:leading_edge:x:local"]
        y3_wing = inputs["data:geometry:wing:kink:y"]
        l4_wing = inputs["data:geometry:wing:tip:chord"]
        x4_wing = inputs["data:geometry:wing:tip:leading_edge:x:local"]
        y4_wing = inputs["data:geometry:wing:tip:y"]

        if y3_wing <= y2_wing:
            y_values = np.squeeze([y2_wing, y4_wing])
            x_values = np.squeeze([[0.0], x4_wing])
            l_values = np.squeeze([l2_wing, l4_wing])
        else:
            y_values = np.squeeze([y2_wing, y3_wing, y4_wing])
            x_values = np.squeeze([[0.0], x3_wing, x4_wing])
            l_values = np.squeeze([l2_wing, l3_wing, l4_wing])

        # Linear interpolation to evaluate the central chord
        line_x_leading_edge = make_interp_spline(y_values, x_values, k=1)
        line_l_center = make_interp_spline(y_values, l_values, k=1)

        outputs["data:geometry:wing:center:chord"] = line_l_center(0.0)
        outputs["data:geometry:wing:center:leading_edge:x:local"] = line_x_leading_edge(0.0)



"""
Estimation of fuel lines weight
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import fastoad.api as oad
import numpy as np
import openmdao.api as om

from .constants import SERVICE_UNCONSUMABLES_MASS


@oad.RegisterSubmodel(
    SERVICE_UNCONSUMABLES_MASS, "fastoad.submodel.weight.mass.propulsion.unconsumables.legacy"
)
class UnconsumablesWeight(om.ExplicitComponent):
    """
    Weight estimation for oil and unusable fuel

    Based on formula in :cite:`supaero:2014`, mass contribution B3
    """

    def setup(self):
        self.add_input("data:geometry:propulsion:engine:count", val=np.nan)
        self.add_input("data:weight:aircraft:MFW", val=np.nan, units="kg")
        self.add_input("tuning:weight:propulsion:unconsumables:mass:k", val=1.0)
        self.add_input("tuning:weight:propulsion:unconsumables:mass:offset", val=0.0, units="kg")

        self.add_output("data:weight:propulsion:unconsumables:mass", units="kg")

    def setup_partials(self):
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        n_engines = inputs["data:geometry:propulsion:engine:count"]
        mfw = inputs["data:weight:aircraft:MFW"]
        k_b3 = inputs["tuning:weight:propulsion:unconsumables:mass:k"]
        offset_b3 = inputs["tuning:weight:propulsion:unconsumables:mass:offset"]

        temp_b3 = 25 * n_engines + 0.0035 * mfw
        outputs["data:weight:propulsion:unconsumables:mass"] = k_b3 * temp_b3 + offset_b3



"""
Estimation of relative kink from an absolute value
"""

#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2025 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om


class ComputeRelativeKink(om.ExplicitComponent):
    """Estimation of relative kink from an absolute value"""

    def setup(self):
        self.add_input("data:geometry:wing:kink:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:tip:y", val=np.nan, units="m")

        self.add_output("data:geometry:wing:kink:span_ratio")

    def setup_partials(self):
        self.declare_partials(
            "data:geometry:wing:kink:span_ratio",
            [
                "data:geometry:wing:kink:y",
                "data:geometry:wing:tip:y",
            ],
            method="fd",
        )

    def compute(self, inputs, outputs):
        y3_wing = inputs["data:geometry:wing:kink:y"]
        y4_wing = inputs["data:geometry:wing:tip:y"]

        if y4_wing > 0:
            wing_break = y3_wing / y4_wing
        else:
            wing_break = 0.0

        outputs["data:geometry:wing:kink:span_ratio"] = wing_break



"""Python package for evaluating aerostructural loads."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2025 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import fastoad.api as oad
import openmdao.api as om
from fastoad.module_management.constants import ModelDomain

from .constants import (
    SERVICE_GUST_LOADS,
    SERVICE_MANEUVER_LOADS,
    SERVICE_SIZING_LOADS_ENVELOPE,
    SERVICE_SIZING_LOADS_MAX,
)


@oad.RegisterOpenMDAOSystem("fastoad.loads.legacy", domain=ModelDomain.LOADS)
class ComputeLoads(om.Group):
    def initialize(self):
        self.options.declare(
            "fuel_load_alleviation",
            types=bool,
            default=True,
            desc="If False this simulates a dry wing,"
            "i.e. the sizing load 2 (pull-up/vertical gust at MTOW) does not take into account "
            "the fuel weight.",
        )

    def setup(self):
        fuel_load_alleviation_option = {
            "fuel_load_alleviation": self.options["fuel_load_alleviation"]
        }
        self.add_subsystem(
            "gust_loads",
            oad.RegisterSubmodel.get_submodel(SERVICE_GUST_LOADS),
            promotes=["*"],
        )
        self.add_subsystem(
            "maneuver_loads",
            oad.RegisterSubmodel.get_submodel(SERVICE_MANEUVER_LOADS),
            promotes=["*"],
        )
        self.add_subsystem(
            "sizing_loads_envelope",
            oad.RegisterSubmodel.get_submodel(
                SERVICE_SIZING_LOADS_ENVELOPE, fuel_load_alleviation_option
            ),
            promotes=["*"],
        )
        self.add_subsystem(
            "sizing_loads_max",
            oad.RegisterSubmodel.get_submodel(SERVICE_SIZING_LOADS_MAX),
            promotes=["*"],
        )



"""
Python module for the computation of the maximum sizing loads.
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2025 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import fastoad.api as oad
import numpy as np
import openmdao.api as om

from ..constants import SERVICE_SIZING_LOADS_MAX


@oad.RegisterSubmodel(SERVICE_SIZING_LOADS_MAX, "fastoad.submodel.loads.maximum_sizing.legacy")
class SizingLoadsEnvelope(om.ExplicitComponent):
    """
    Computes CS25 sizing loading

    """

    def setup(self):
        self.add_input("data:mission:sizing:cs25:envelope:max_sizing_load_1", units="N")
        self.add_input("data:mission:sizing:cs25:envelope:max_sizing_load_2", units="N")
        self.add_input("data:mission:sizing:cs25:envelope:max_load_factor_1", units="unitless")
        self.add_input("data:mission:sizing:cs25:envelope:max_load_factor_2", units="unitless")

        self.add_output("data:mission:sizing:cs25:sizing_load", units="N")
        self.add_output("data:mission:sizing:cs25:sizing_load_factor", units="unitless")

    def setup_partials(self):
        self.declare_partials(
            "data:mission:sizing:cs25:sizing_load",
            "*",
            method="fd",
        )

        self.declare_partials(
            "data:mission:sizing:cs25:sizing_load_factor",
            "*",
            method="fd",
        )

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        n1m1 = inputs["data:mission:sizing:cs25:envelope:max_sizing_load_1"]
        n2m2 = inputs["data:mission:sizing:cs25:envelope:max_sizing_load_2"]
        n1 = inputs["data:mission:sizing:cs25:envelope:max_load_factor_1"]
        n2 = inputs["data:mission:sizing:cs25:envelope:max_load_factor_2"]

        id_max = int(np.argmax([n1m1, n2m2]))
        nm = [n1m1, n2m2][id_max]
        n = [n1, n2][id_max]

        outputs["data:mission:sizing:cs25:sizing_load"] = nm
        outputs["data:mission:sizing:cs25:sizing_load_factor"] = n



"""Computation of furniture mass."""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import fastoad.api as oad
from openmdao import api as om

from .constants import (
    SERVICE_FOOD_WATER_MASS,
    SERVICE_PASSENGER_SEATS_MASS,
    SERVICE_SECURITY_KIT_MASS,
    SERVICE_TOILETS_MASS,
)
from ..constants import SERVICE_FURNITURE_MASS


@oad.RegisterSubmodel(
    SERVICE_FURNITURE_MASS, "fastoad.submodel.weight.mass.furniture.cargo_configuration.legacy"
)
class FurnitureWeight(om.Group):
    """
    Computes mass of furniture.
    """

    def setup(self):
        self.add_subsystem(
            "passenger_seats_weight",
            oad.RegisterSubmodel.get_submodel(SERVICE_PASSENGER_SEATS_MASS),
            promotes=["*"],
        )
        self.add_subsystem(
            "food_water_weight",
            oad.RegisterSubmodel.get_submodel(SERVICE_FOOD_WATER_MASS),
            promotes=["*"],
        )
        self.add_subsystem(
            "security_kit_weight",
            oad.RegisterSubmodel.get_submodel(SERVICE_SECURITY_KIT_MASS),
            promotes=["*"],
        )
        self.add_subsystem(
            "toilets_weight",
            oad.RegisterSubmodel.get_submodel(SERVICE_TOILETS_MASS),
            promotes=["*"],
        )

        weight_sum = om.AddSubtractComp()
        weight_sum.add_equation(
            "data:weight:furniture:mass",
            [
                "data:weight:furniture:passenger_seats:mass",
                "data:weight:furniture:food_water:mass",
                "data:weight:furniture:security_kit:mass",
                "data:weight:furniture:toilets:mass",
            ],
            units="kg",
            desc="Mass of aircraft furniture",
        )

        self.add_subsystem("furniture_weight_sum", weight_sum, promotes=["*"])



"""
Python module for the computation of maneuver and gust sizing load case envelope.
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2025 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import fastoad.api as oad
import numpy as np
import openmdao.api as om
from scipy.constants import g

from ..constants import SERVICE_SIZING_LOADS_ENVELOPE


@oad.RegisterSubmodel(SERVICE_SIZING_LOADS_ENVELOPE, "fastoad.submodel.loads.envelope.legacy")
class SizingLoadsEnvelope(om.ExplicitComponent):
    """
    Computes CS25 sizing loads evaluated at two different load cases:

    Load case 1: with wings with almost no fuel
    Load case 2: at maximum take-off weight

    Based on formulas in :cite:`supaero:2014`, §6.3.
    """

    def initialize(self):
        self.options.declare(
            "fuel_load_alleviation",
            types=bool,
            default=True,
            desc="If False this simulates a dry wing,"
            "i.e. the sizing load 2 does not take into account the fuel weight.",
        )

    def setup(self):
        self.add_input("data:weight:aircraft:MZFW", val=np.nan, units="kg")
        self.add_input("data:weight:aircraft:MFW", val=np.nan, units="kg")
        self.add_input("data:weight:aircraft:MTOW", val=np.nan, units="kg")
        self.add_input("data:mission:sizing:cs25:gust:load_factor_1", units="unitless")
        self.add_input("data:mission:sizing:cs25:gust:load_factor_2", units="unitless")
        self.add_input("data:mission:sizing:cs25:maneuver:load_factor_1", units="unitless")
        self.add_input("data:mission:sizing:cs25:maneuver:load_factor_2", units="unitless")

        self.add_output("data:mission:sizing:cs25:envelope:max_load_factor_1", units="unitless")
        self.add_output("data:mission:sizing:cs25:envelope:max_load_factor_2", units="unitless")
        self.add_output("data:mission:sizing:cs25:envelope:max_sizing_load_1", units="N")
        self.add_output("data:mission:sizing:cs25:envelope:max_sizing_load_2", units="N")

    def setup_partials(self):
        # Load factor outputs
        self.declare_partials(
            "data:mission:sizing:cs25:envelope:max_load_factor_1",
            [
                "data:mission:sizing:cs25:gust:load_factor_1",
                "data:mission:sizing:cs25:maneuver:load_factor_1",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:mission:sizing:cs25:envelope:max_load_factor_2",
            [
                "data:mission:sizing:cs25:gust:load_factor_2",
                "data:mission:sizing:cs25:maneuver:load_factor_2",
            ],
            method="fd",
        )

        # Sizing load outputs
        self.declare_partials(
            "data:mission:sizing:cs25:envelope:max_sizing_load_1",
            [
                "data:mission:sizing:cs25:gust:load_factor_1",
                "data:mission:sizing:cs25:maneuver:load_factor_1",
                "data:weight:aircraft:MZFW",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:mission:sizing:cs25:envelope:max_sizing_load_2",
            [
                "data:mission:sizing:cs25:gust:load_factor_2",
                "data:mission:sizing:cs25:maneuver:load_factor_2",
                "data:weight:aircraft:MFW",
                "data:weight:aircraft:MTOW",
            ],
            method="fd",
        )

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        mzfw = inputs["data:weight:aircraft:MZFW"]
        mfw = inputs["data:weight:aircraft:MFW"]
        mtow = inputs["data:weight:aircraft:MTOW"]

        n1_maneuver = inputs["data:mission:sizing:cs25:maneuver:load_factor_1"]
        n2_maneuver = inputs["data:mission:sizing:cs25:maneuver:load_factor_2"]
        n1_gust = inputs["data:mission:sizing:cs25:gust:load_factor_1"]
        n2_gust = inputs["data:mission:sizing:cs25:gust:load_factor_2"]

        # load case #1
        m1 = 1.05 * mzfw

        # load case #2
        if not self.options["fuel_load_alleviation"]:
            m2 = mtow
        else:
            mcv = min(0.8 * mfw, mtow - mzfw)  # fuel mass in the overhanging part of wing
            m2 = mtow - 0.55 * mcv

        n1 = max(n1_gust, n1_maneuver)
        n2 = max(n2_gust, n2_maneuver)

        n1m1 = n1 * m1 * g
        n2m2 = n2 * m2 * g

        outputs["data:mission:sizing:cs25:envelope:max_load_factor_1"] = n1
        outputs["data:mission:sizing:cs25:envelope:max_load_factor_2"] = n2
        outputs["data:mission:sizing:cs25:envelope:max_sizing_load_1"] = n1m1
        outputs["data:mission:sizing:cs25:envelope:max_sizing_load_2"] = n2m2



"""
Python module for the computation of gust sizing load cases.
"""
#  This file is part of FAST-OAD_CS25
#  Copyright (C) 2025 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import fastoad.api as oad
import numpy as np
import openmdao.api as om
from scipy.constants import g
from stdatm import Atmosphere

from ..constants import SERVICE_GUST_LOADS


@oad.RegisterSubmodel(SERVICE_GUST_LOADS, "fastoad.submodel.loads.gust.legacy")
class GustLoads(om.ExplicitComponent):
    """
    Computes CS25 vertical gust load factors evaluated at two different load cases:

    Load case 1: with wings with almost no fuel
    Load case 2: at maximum take-off weight

    Based on formulas in :cite:`supaero:2014`, §6.3

    """

    def setup(self):
        self.add_input("data:geometry:wing:area", val=np.nan, units="m**2")
        self.add_input("data:geometry:wing:span", val=np.nan, units="m")
        self.add_input("data:weight:aircraft:MZFW", val=np.nan, units="kg")
        self.add_input("data:weight:aircraft:MTOW", val=np.nan, units="kg")
        self.add_input("data:aerodynamics:aircraft:cruise:CL_alpha", val=np.nan, units="1/rad")
        self.add_input("data:load_case:lc1:U_gust", val=np.nan, units="m/s")
        self.add_input("data:load_case:lc1:altitude", val=np.nan, units="ft")
        self.add_input("data:load_case:lc1:Vc_EAS", val=np.nan, units="m/s")
        self.add_input("data:load_case:lc2:U_gust", val=np.nan, units="m/s")
        self.add_input("data:load_case:lc2:altitude", val=np.nan, units="ft")
        self.add_input("data:load_case:lc2:Vc_EAS", val=np.nan, units="m/s")
        self.add_input("data:load_case:gust_intensity", val=1.0, units="unitless")
        self.add_input("data:mission:sizing:cs25:safety_factor", val=1.5, units="unitless")

        self.add_output("data:mission:sizing:cs25:gust:load_factor_1", units="unitless")
        self.add_output("data:mission:sizing:cs25:gust:load_factor_2", units="unitless")

    def setup_partials(self):
        self.declare_partials(
            "data:mission:sizing:cs25:gust:load_factor_1",
            [
                "data:geometry:wing:area",
                "data:geometry:wing:span",
                "data:weight:aircraft:MZFW",
                "data:aerodynamics:aircraft:cruise:CL_alpha",
                "data:load_case:lc1:U_gust",
                "data:load_case:lc1:altitude",
                "data:load_case:lc1:Vc_EAS",
                "data:load_case:gust_intensity",
                "data:mission:sizing:cs25:safety_factor",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:mission:sizing:cs25:gust:load_factor_2",
            [
                "data:geometry:wing:area",
                "data:geometry:wing:span",
                "data:weight:aircraft:MTOW",
                "data:aerodynamics:aircraft:cruise:CL_alpha",
                "data:load_case:lc2:U_gust",
                "data:load_case:lc2:altitude",
                "data:load_case:lc2:Vc_EAS",
                "data:load_case:gust_intensity",
                "data:mission:sizing:cs25:safety_factor",
            ],
            method="fd",
        )

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        sea_level_density = Atmosphere(0).density
        wing_area = inputs["data:geometry:wing:area"]
        span = inputs["data:geometry:wing:span"]
        mzfw = inputs["data:weight:aircraft:MZFW"]
        mtow = inputs["data:weight:aircraft:MTOW"]
        cl_alpha = inputs["data:aerodynamics:aircraft:cruise:CL_alpha"]
        u_gust1 = inputs["data:load_case:lc1:U_gust"]
        alt_1 = inputs["data:load_case:lc1:altitude"]
        vc_eas1 = inputs["data:load_case:lc1:Vc_EAS"]
        u_gust2 = inputs["data:load_case:lc2:U_gust"]
        alt_2 = inputs["data:load_case:lc2:altitude"]
        vc_eas2 = inputs["data:load_case:lc2:Vc_EAS"]
        gust_intensity = inputs["data:load_case:gust_intensity"]
        safety_factor = inputs["data:mission:sizing:cs25:safety_factor"]

        # calculation of mean geometric chord
        chord_geom = wing_area / span

        # load case #1
        m1 = 1.05 * mzfw
        n_gust_1 = self.__n_gust(
            m1,
            wing_area,
            Atmosphere(alt_1).density,
            sea_level_density,
            chord_geom,
            vc_eas1,
            cl_alpha,
            u_gust1,
        )
        n1 = safety_factor * n_gust_1 * gust_intensity

        # load case #2
        n_gust_2 = self.__n_gust(
            mtow,
            wing_area,
            Atmosphere(alt_2).density,
            sea_level_density,
            chord_geom,
            vc_eas2,
            cl_alpha,
            u_gust2,
        )
        n2 = safety_factor * n_gust_2 * gust_intensity

        outputs["data:mission:sizing:cs25:gust:load_factor_1"] = n1
        outputs["data:mission:sizing:cs25:gust:load_factor_2"] = n2

    @staticmethod
    def __n_gust(mass, wing_area, rho, sea_level_density, chord_geom, vc_eas, cl_alpha, u_gust):
        """
        Computes a reference vertical gust load factor.

        :param mass: Aircraft mass [kg]
        :param wing_area: Wing reference area [m2]
        :param rho: Air density at flight altitude [kg/m3]
        :param sea_level_density: Air density at sea level [kg/m3]
        :param chord_geom: Geometric mean aerodynamic chord [m]
        :param vc_eas: Equivalent airspeed at cruising condition Vc [m/s]
        :param cl_alpha: Wing lift alpha curve slope [1/rad]
        :param u_gust: Gust vertical velocity [m/s]
        :return: Gust load factor (n_gust) [dimensionless]
        """
        mu_g = 2 * mass / rho / wing_area / chord_geom / cl_alpha
        k_g = 0.88 * mu_g / (5.3 + mu_g)  # attenuation factor
        return 1 + (sea_level_density / 2 / g) * k_g * u_gust * (
            vc_eas * cl_alpha / (mass / wing_area)
        )  # n_gust
